#include "ruby.h"
#include <env.h>

#include <stdio.h>
#include <string.h>

#include "gpmi/debug.h"

static int ruby_module_loaded = 0;

static VALUE Ruby_DynamicPackageCommand(int argc, VALUE *argv)
{
	char *func_name = rb_id2name(ruby_frame->orig_func);
	char *(*proc)(int argc, char *argv[]);
	char **myargv;
	VALUE ruby_res;
	char *res;
	int i;

	/* Find our proc based on the function-name */
	proc = GPMI_Script_FindProc(func_name);

	if (proc == NULL) {
		gpmi_error(0, "[Ruby] Internal error: %s has a NULL proc in the command hash\n", func_name);
		return Qnil;
	}

	myargv = malloc((argc + 1) * sizeof(char *));

	myargv[0] = func_name;

	/* Make sure all args are strings */
	for (i = 0; i < argc; i++) {
		switch (rb_type(argv[i])) {
			/* String */
			case T_STRING:
				myargv[i + 1] = strdup(RSTRING(argv[i])->ptr);
				break;

			/* Integer */
			case T_FIXNUM: {
				char tmp[100];
				snprintf(tmp, sizeof(tmp), "%d", (int)NUM2INT(argv[i]));
				myargv[i + 1] = strdup(tmp);
			} break;

			default:
				/* NOT SUPPORTED! */
				gpmi_error(0, "[Ruby] Internal error: %d is an unsupported parameter\n", rb_type(argv[i]));
				return Qnil;
		}
	}

	/* Call the function */
	res = proc(argc + 1, myargv);

	/* Free the memory */
	for (i = 1; i < argc; i++)
		free(myargv[i]);
	free(myargv);

	/* Find the right return value and return */
	if (res == NULL)
		return Qnil;

	ruby_res = rb_str_new2(res);
	free(res);

	return ruby_res;
}

void Ruby_RegisterCommand(void *client_data, const char *name)
{
	/* Simple, easy, works */
	rb_define_global_function(name, &Ruby_DynamicPackageCommand, -1);
}

int Ruby_LoadScript(void *client_data, const char *filename)
{
	/* Load the file, and execute once */
	rb_load_file(filename);
	ruby_exec();

	if (ruby_errinfo != Qnil)
		gpmi_error(1, "[Ruby] Error (%s:%d): %s", ruby_sourcefile, ruby_sourceline,  RSTRING(rb_obj_as_string(ruby_errinfo))->ptr);

	return 0;
}

static VALUE Ruby_ExceptionHandler(void)
{
	/* An error, show it */
	VALUE lasterr = rb_gv_get("$!");
	VALUE message = rb_obj_as_string(lasterr);

	gpmi_error(1, "[Ruby] Warning (%s:%d): %s", ruby_sourcefile, ruby_sourceline, RSTRING(message)->ptr);

	return Qnil;
}

static VALUE Ruby_SafeCallFunction(VALUE *ar)
{
	/* Call the function safely */
	rb_funcall2(ar[0], ar[1], ar[2], (VALUE *)ar[3]);

	return Qnil;
}

void Ruby_CallFunction(void *client_data, const char *function, int argc, const char **argv)
{
	VALUE ar[4];
	VALUE *new_argv;
	int i;

	new_argv = malloc(sizeof(VALUE) * argc);

	for (i = 0; i < argc; i++)
		new_argv[i] = rb_str_new2(argv[i]);

	/* Put data in array */
	ar[0] = ruby_errinfo;
	ar[1] = rb_intern(function);
	ar[2] = argc;
	ar[3] = (VALUE)new_argv;

	/* Call function in safe-mode */
	rb_rescue(Ruby_SafeCallFunction, (VALUE)ar, Ruby_ExceptionHandler, 0);

	free(new_argv);
}

int Ruby_Initialize(void **client_data)
{
	if (ruby_module_loaded >= 1) {
		gpmi_error(0, "[Ruby] This module can only have one active instance");
		return -1;
	}


	ruby_init();
	ruby_script("gpmi");
	ruby_init_loadpath();

	ruby_module_loaded++;

	return 0;
}

void Ruby_Destroy(void **client_data)
{
	ruby_finalize();

	ruby_module_loaded--;
}

char *Ruby_GetExtension()
{
	return strdup(".rb");
}
