#include <stdlib.h>
#include <string.h>
#include "module_events.h"

module_events *module_events_init()
{
	module_events *m;

	m = malloc(sizeof(module_events));

	m->allocated = 0;
	m->events    = NULL;

	return m;
}

int module_events_uninit(module_events *events, module_free_data_proc *proc)
{
	int i;
	module_event *e, *next;

	/* Clean up the linked-list */
	for (i = 0; i < events->allocated; i++) {
		e = events->events[i];

		while (e != NULL) {
			next = e->next;
			proc(e->data);
			free(e);
			e = next;
		}
	}

	/* Clean up the array and struct */
	free(events->events);
	free(events);

	return 0;
}

void module_events_bind(module_events *events, int event_id, void *event_data)
{
	module_event *e;
	int old_len;

	/* Exapand the list array if needed */
	if (event_id >= events->allocated) {
		old_len = events->allocated;
		/* Make 16 more free slots than needed */
		events->allocated = event_id + 16;

		events->events    = realloc(events->events, events->allocated * sizeof(module_event *));
		/* clean newly allocated memory */
		memset(events->events + old_len, 0, (events->allocated - old_len) * sizeof(module_event *));
	}

	/* Prepare the new event's data */
	e = malloc(sizeof(module_event));
	e->data = event_data;
	/* Make sure we don't call a bind that is binded inside an event-call */
	e->called = 2;

	/* Add an entry in the linked list: insert it before any other events */
	e->next = events->events[event_id];
	events->events[event_id] = e;
}


int module_events_unbind(module_events *events, int event_id, void *event_data, module_compare_data_proc *proc_match, module_free_data_proc *proc)
{
	module_event *e, *last;

	/* Search the whole linked-list for event_data */
	for (e = events->events[event_id], last = NULL; e != NULL; last = e, e = e->next) {
		if (proc_match(e->data, event_data)) {
			/* Correct the linked-list */
			if (last != NULL)
				last->next = e->next;
			else
				events->events[event_id] = e->next;

			proc(e->data);

			/* Free the item */
			free(e);

			return 0;
		}
	}

	return -1;
}

int module_events_call(module_events *events, int event_id, module_events_callback *callback, void *handle)
{
	module_event *e;

	/* Out of the event array? */
	if (event_id >= events->allocated)
		return -1;

	e = events->events[event_id];
	if (e == NULL)
		return -2;

	if (e->next == NULL) {
		callback(handle, e->data);
		return 0;
	}

	/* Mark all as non-called */
	for(;e != NULL; e = e->next) {
		e->called = 0;
	}
	/* Go back */
	e = events->events[event_id];

	/* Scroll through all events, go back after each time, and find a new un-called.. this to avoid
	 *  memory problems because a callback can call an unbind, making the linked-list invalid
	 *  from our current point. */
	for(;e != NULL; e = e->next) {
		if (e->called != 0)
			continue;

		/* Call the proc */
		e->called = 1;
		callback(handle, e->data);

		/* Go back to the first item */
		e = events->events[event_id];
	}

	return 0;
}
